
/* test_subscriber.cxx

A subscription example

This file is derived from code automatically generated by the rtiddsgen
command:

rtiddsgen -language C++ -example <arch> test.idl

Example subscription of type test automatically generated by
'rtiddsgen'. To test them follow these steps:

(1) Compile this file and the example publication.

(2) Start the subscription with the command
objs/<arch>/test_subscriber <domain_id> <sample_count>

(3) Start the publication with the command
objs/<arch>/test_publisher <domain_id> <sample_count>

(4) [Optional] Specify the list of discovery initial peers and
multicast receive addresses via an environment variable or a file
(in the current working directory) called NDDS_DISCOVERY_PEERS.

You can run any number of publishers and subscribers programs, and can
add and remove them dynamically from the domain.

Example:

To run the example application on domain <domain_id>:

On Unix:

objs/<arch>/test_publisher <domain_id>
objs/<arch>/test_subscriber <domain_id>

On Windows:

objs\<arch>\test_publisher <domain_id>
objs\<arch>\test_subscriber <domain_id>

*/

#include <stdio.h>
#include <stdlib.h>
#include <map>
#include <vector>
#include <string>

#include "test.h"
#include "testSupport.h"
#include "ndds/ndds_cpp.h"

std::map<DDS_GUID_t, std::string> register_topic;

/** Taken from http://community.rti.com/comment/689#comment-689 */
DDS_GUID_t convertToGuid ( const DDS_InstanceHandle_t& instanceHandle ) {
    DDS_GUID_t guid;
    memcpy(guid.value,reinterpret_cast<DDS_Octet const *>(&instanceHandle ),16);
    return guid;
}
/** Taken from http://community.rti.com/rti-doc/500/ndds.5.0.0/doc/html/api_cpp/group__DDSGUIDSupportModule.html# */
inline bool operator==(DDS_GUID_t const & lhs, DDS_GUID_t const & rhs) { return DDS_BOOLEAN_TRUE == DDS_GUID_equals(&lhs,&rhs);};
inline bool operator!=(DDS_GUID_t const & lhs, DDS_GUID_t const & rhs) { return !operator == (lhs,rhs);};
inline bool operator< (DDS_GUID_t const & lhs, DDS_GUID_t const & rhs) { return DDS_GUID_compare(&lhs,&rhs) < 0;};
inline bool operator> (DDS_GUID_t const & lhs, DDS_GUID_t const & rhs) { return  operator< (rhs,lhs);};
inline bool operator<=(DDS_GUID_t const & lhs, DDS_GUID_t const & rhs) { return !operator> (lhs,rhs);};
inline bool operator>=(DDS_GUID_t const & lhs, DDS_GUID_t const & rhs) { return !operator< (lhs,rhs);};

class BuiltinPublicationListener : public DDSDataReaderListener {

public:
    virtual void on_data_available(DDSDataReader* reader);
};

/* This gets called when a new subscriber has been discovered */
void BuiltinPublicationListener::on_data_available(DDSDataReader *reader)
{
    DDSPublicationBuiltinTopicDataDataReader *builtin_reader =
        (DDSPublicationBuiltinTopicDataDataReader*) reader;
    DDS_PublicationBuiltinTopicDataSeq data_seq;
    DDS_SampleInfoSeq info_seq;
    DDS_ReturnCode_t retcode;

    /* We only process newly seen subscribers */
    retcode = builtin_reader->take(data_seq, info_seq, DDS_LENGTH_UNLIMITED,
            DDS_ANY_SAMPLE_STATE, DDS_ANY_VIEW_STATE, DDS_ANY_INSTANCE_STATE);

    if (retcode == DDS_RETCODE_NO_DATA)
        return;

    if (retcode != DDS_RETCODE_OK) {
        printf("***Error: failed to access data from the built-in reader\n");
        return;
    }
    for (int i = 0; i < data_seq.length(); ++i) {
        switch (info_seq[i].instance_state) {
            case DDS_ALIVE_INSTANCE_STATE:
                printf("__Built-in : found DW \n");
                break;
            case DDS_NOT_ALIVE_DISPOSED_INSTANCE_STATE:
                printf("__Built-in : lost DW, gracefully terminated \n");
                break;
            case DDS_NOT_ALIVE_NO_WRITERS_INSTANCE_STATE:
                printf("__Built-in : lost DW, ungracefully terminated \n");
                break;
        } //End Case for instance state
        if (info_seq[i].valid_data) {
            printf("\tTopic:     %s\n", data_seq[i].topic_name);
            register_topic[convertToGuid(info_seq[i].instance_handle)] =
                    data_seq[i].topic_name;
        } else {
            printf("\tTopic:     %s\n",
                    register_topic[convertToGuid(info_seq[i].instance_handle)].c_str());
        }
    }
    builtin_reader->return_loan(data_seq, info_seq);
}

class testListener : public DDSDataReaderListener {
  public:
    virtual void on_requested_deadline_missed(
        DDSDataReader* /*reader*/,
        const DDS_RequestedDeadlineMissedStatus& /*status*/) {}

    virtual void on_requested_incompatible_qos(
        DDSDataReader* /*reader*/,
        const DDS_RequestedIncompatibleQosStatus& /*status*/) {}

    virtual void on_sample_rejected(
        DDSDataReader* /*reader*/,
        const DDS_SampleRejectedStatus& /*status*/) {}

    virtual void on_liveliness_changed(
        DDSDataReader* /*reader*/,
        const DDS_LivelinessChangedStatus& /*status*/) {}

    virtual void on_sample_lost(
        DDSDataReader* /*reader*/,
        const DDS_SampleLostStatus& /*status*/) {}

    virtual void on_subscription_matched(
        DDSDataReader* /*reader*/,
        const DDS_SubscriptionMatchedStatus& /*status*/) {}

    virtual void on_data_available(DDSDataReader* reader);
};

void testListener::on_data_available(DDSDataReader* reader)
{
    testDataReader *test_reader = NULL;
    testSeq data_seq;
    DDS_SampleInfoSeq info_seq;
    DDS_ReturnCode_t retcode;
    int i;

    test_reader = testDataReader::narrow(reader);
    if (test_reader == NULL) {
        fprintf(stderr, "DataReader narrow error\n");
        return;
    }

    retcode = test_reader->take(
        data_seq, info_seq, DDS_LENGTH_UNLIMITED,
        DDS_ANY_SAMPLE_STATE, DDS_ANY_VIEW_STATE, DDS_ANY_INSTANCE_STATE);

    if (retcode == DDS_RETCODE_NO_DATA) {
        return;
    } else if (retcode != DDS_RETCODE_OK) {
        fprintf(stderr, "take error %d\n", retcode);
        return;
    }

    for (i = 0; i < data_seq.length(); ++i) {
        if (info_seq[i].valid_data) {
            printf("Received data\n");
            testTypeSupport::print_data(&data_seq[i]);
        }
    }

    retcode = test_reader->return_loan(data_seq, info_seq);
    if (retcode != DDS_RETCODE_OK) {
        fprintf(stderr, "return loan error %d\n", retcode);
    }
}

/* Delete all entities */
static int subscriber_shutdown(
    DDSDomainParticipant *participant)
{
    DDS_ReturnCode_t retcode;
    int status = 0;

    if (participant != NULL) {
        retcode = participant->delete_contained_entities();
        if (retcode != DDS_RETCODE_OK) {
            fprintf(stderr, "delete_contained_entities error %d\n", retcode);
            status = -1;
        }

        retcode = DDSTheParticipantFactory->delete_participant(participant);
        if (retcode != DDS_RETCODE_OK) {
            fprintf(stderr, "delete_participant error %d\n", retcode);
            status = -1;
        }
    }

    /* RTI Connext provides the finalize_instance() method on
    domain participant factory for people who want to release memory used
    by the participant factory. Uncomment the following block of code for
    clean destruction of the singleton. */
    /*

    retcode = DDSDomainParticipantFactory::finalize_instance();
    if (retcode != DDS_RETCODE_OK) {
        fprintf(stderr, "finalize_instance error %d\n", retcode);
        status = -1;
    }
    */
    return status;
}

extern "C" int subscriber_main(int domainId, int sample_count)
{
    DDSDomainParticipant *participant = NULL;
    DDSSubscriber *subscriber = NULL;
    DDSTopic *topic = NULL;
    testListener *reader_listener = NULL;
    DDSDataReader *reader = NULL;
    DDS_ReturnCode_t retcode;
    const char *type_name = NULL;
    int count = 0;
    DDS_Duration_t receive_period = {4,0};
    int status = 0;

    /* To customize the participant QoS, use
    the configuration file USER_QOS_PROFILES.xml */
    participant = DDSTheParticipantFactory->create_participant(
        domainId, DDS_PARTICIPANT_QOS_DEFAULT,
        NULL /* listener */, DDS_STATUS_MASK_NONE);
    if (participant == NULL) {
        fprintf(stderr, "create_participant error\n");
        subscriber_shutdown(participant);
        return -1;
    }

    /* Start changes for Builtin_Topics */
    /* First get the builtin subscriber */
    DDSSubscriber *builtin_subscriber = participant->get_builtin_subscriber();
    if (builtin_subscriber == NULL) {
        printf("***Error: failed to create builtin subscriber\n");
        subscriber_shutdown(participant);
        return 0;
    }

    /* Get builtin subscriber's datareader for publisher */
    DDSPublicationBuiltinTopicDataDataReader *builtin_publication_datareader =
            (DDSPublicationBuiltinTopicDataDataReader*)
            builtin_subscriber->lookup_datareader(DDS_PUBLICATION_TOPIC_NAME);
    if (builtin_publication_datareader == NULL) {
      printf("***Error: failed to create builtin subscription data reader\n");
      return 0;
    }

    BuiltinPublicationListener *builtin_publication_listener =
            new BuiltinPublicationListener();
    builtin_publication_datareader->set_listener(builtin_publication_listener,
                                                DDS_STATUS_MASK_ALL);
    /* Done!  All the listeners are installed, so we can enable the
     * participant now.
     */
    if (participant->enable() != DDS_RETCODE_OK) {
        printf("***Error: Failed to Enable Participant\n");
        subscriber_shutdown(participant);
        delete builtin_publication_listener;
        return 0;
    }
    /* End changes for Builtin_Topics */

    /* To customize the subscriber QoS, use
    the configuration file USER_QOS_PROFILES.xml */
    subscriber = participant->create_subscriber(
            DDS_SUBSCRIBER_QOS_DEFAULT, NULL /* listener */, DDS_STATUS_MASK_NONE);
    if (subscriber == NULL) {
        fprintf(stderr, "create_subscriber error\n");
        subscriber_shutdown(participant);
        delete builtin_publication_listener;
        return -1;
    }

    /* Register the type before creating the topic */
    type_name = testTypeSupport::get_type_name();
    retcode = testTypeSupport::register_type(
        participant, type_name);
    if (retcode != DDS_RETCODE_OK) {
        fprintf(stderr, "register_type error %d\n", retcode);
        subscriber_shutdown(participant);
        delete builtin_publication_listener;
        return -1;
    }

    /* To customize the topic QoS, use
    the configuration file USER_QOS_PROFILES.xml */
    topic = participant->create_topic(
        "Example test",
        type_name, DDS_TOPIC_QOS_DEFAULT, NULL /* listener */,
        DDS_STATUS_MASK_NONE);
    if (topic == NULL) {
        fprintf(stderr, "create_topic error\n");
        subscriber_shutdown(participant);
        delete builtin_publication_listener;
        return -1;
    }

    /* Create a data reader listener */
    reader_listener = new testListener();

    /* To customize the data reader QoS, use
    the configuration file USER_QOS_PROFILES.xml */
    reader = subscriber->create_datareader(
        topic, DDS_DATAREADER_QOS_DEFAULT, reader_listener,
        DDS_STATUS_MASK_ALL);
    if (reader == NULL) {
        fprintf(stderr, "create_datareader error\n");
        subscriber_shutdown(participant);
        delete builtin_publication_listener;
        delete reader_listener;
        return -1;
    }

    /* Main loop */
    for (count=0; (sample_count == 0) || (count < sample_count); ++count) {

        printf("test subscriber sleeping for %d sec...\n",
        receive_period.sec);

        NDDSUtility::sleep(receive_period);
    }

    /* Delete all entities */
    status = subscriber_shutdown(participant);
    delete builtin_publication_listener;
    delete reader_listener;

    return status;
}

int main(int argc, char *argv[])
{
    int domainId = 0;
    int sample_count = 0; /* infinite loop */

    if (argc >= 2) {
        domainId = atoi(argv[1]);
    }
    if (argc >= 3) {
        sample_count = atoi(argv[2]);
    }

    /* Uncomment this to turn on additional logging
    NDDSConfigLogger::get_instance()->
    set_verbosity_by_category(NDDS_CONFIG_LOG_CATEGORY_API,
    NDDS_CONFIG_LOG_VERBOSITY_STATUS_ALL);
    */

    return subscriber_main(domainId, sample_count);
}
